"""Amplitude config module.

Classes:
    Config: Class that store configurations of Amplitude client
"""

import logging
from typing import Optional, Callable

from fga import constants
from fga.event import BaseEvent, Plan, IngestionMetadata
from fga.storage import InMemoryStorageProvider, StorageProvider, Storage


class Config:
    """Amplitude client configuration class used to config client behavior

    Args:
        api_key (str): The API key of Amplitude project that events sent to. Required
        flush_queue_size (int, optional): The events buffered in memory will flush when exceed flush_queue_size.
            Must be positive.
        flush_interval_millis (int, optional): The events buffered in memory will wait no longer than
            flush_interval_millis. Must be positive.
        flush_max_retries (int, optional): The maximum retry attempts for an event when receiving error response.
        logger (optional): The logger used by Amplitude client. Default to logging.getLogger(constants.LOGGER_NAME).
        min_id_length (int, optional): The minimum length of user_id and device_id for events. Default to 5.
        callback (callable, optional): The client level callback function. Triggered on every events sent or failed.
            Take three parameters: an event instance, an integer code of response status, an optional string message.
        server_zone (str, optional): The server zone of project. Default to 'US'. Support 'EU'.
        use_batch(bool, optional): True to use batch API endpoint, False to use HTTP V2 API endpoint. Default to False.
        server_url (str, optional): API endpoint url. Default to None. Auto selected by configured server_zone
            and use_batch if set to None. Support customized url by setting string value.
        storage_provider (amplitude.storage.StorageProvider, optional): Default to InMemoryStorageProvider.
            Provide storage instance for events buffer.
        plan (amplitude.event.Plan, optional): Tracking plan information. Default to None.
        ingestion_metadata (amplitude.event.IngestionMetadata, optional): Ingestion metadata. Default to None.

    Properties:
        options: A dictionary contains minimum id length information. None if min_id_length not set.

    Methods:
        get_storage(): Return a Storage instance using configured StorageProvider.
        is_valid(): Return True if the Config instance is valid, False otherwise.
        is_min_id_length_valid(): Return True if min_id_length not set or set to positive integer, False otherwise.

    """

    def __init__(self, api_key: str = None,
                 project_id: int = None,
                 flush_queue_size: int = constants.FLUSH_QUEUE_SIZE,
                 flush_interval_millis: int = constants.FLUSH_INTERVAL_MILLIS,
                 flush_max_retries: int = constants.FLUSH_MAX_RETRIES,
                 logger=logging.getLogger(constants.LOGGER_NAME),
                 min_id_length: Optional[int] = None,
                 callback: Optional[Callable[[BaseEvent, int, Optional[str]], None]] = None,
                 server_zone: str = constants.DEFAULT_ZONE,
                 use_batch: bool = False,
                 server_url: Optional[str] = None,
                 storage_provider: Optional[StorageProvider] = None,
                 plan: Plan = None,
                 ingestion_metadata: IngestionMetadata = None):
        """The constructor of Config class"""
        self.api_key: str = api_key
        self.project_id: str = project_id
        self._flush_queue_size: int = flush_queue_size
        self._flush_size_divider: int = 1
        self.flush_interval_millis: int = flush_interval_millis
        self.flush_max_retries: int = flush_max_retries
        self.logger = logger
        self.min_id_length: Optional[int] = min_id_length
        self.callback: Optional[Callable[[BaseEvent, int, Optional[str]], None]] = callback
        self.server_zone: str = server_zone
        self.use_batch: bool = use_batch
        self._url: Optional[str] = server_url
        self.storage_provider: StorageProvider = storage_provider or InMemoryStorageProvider()
        self.opt_out: bool = False
        self.plan: Plan = plan
        self.ingestion_metadata: IngestionMetadata = ingestion_metadata

    def get_storage(self) -> Storage:
        """Use configured StorageProvider to create a Storage instance then return.

        Returns:
            A Storage instance
        """
        return self.storage_provider.get_storage()

    def is_valid(self) -> bool:
        """Config instance is valid if api_key is set, flush_queue_size, flush_interval_millis are positive integer,
            min_id_length is valid.

        Returns:
            True if valid. False otherwise.
        """
        if (not self.api_key) or (self.flush_queue_size <= 0) or (
                self.flush_interval_millis <= 0) or (not self.is_min_id_length_valid()):
            return False
        return True

    def is_min_id_length_valid(self) -> bool:
        """min_id_length is valid when set to positive integer or None.

        Returns:
             True if valid. False otherwise.
        """
        if self.min_id_length is None or \
                (isinstance(self.min_id_length, int) and self.min_id_length > 0):
            return True
        return False

    def _increase_flush_divider(self):
        self._flush_size_divider += 1

    def _reset_flush_divider(self):
        self._flush_size_divider = 1

    @property
    def flush_queue_size(self):
        return max(1, self._flush_queue_size // self._flush_size_divider)

    @flush_queue_size.setter
    def flush_queue_size(self, size: int):
        self._flush_queue_size = size
        self._flush_size_divider = 1

    @property
    def server_url(self):
        if self._url:
            return self._url
        return constants.SERVER_URL[constants.EVENT]

    @server_url.setter
    def server_url(self, url: str):
        self._url = url

    @property
    def options(self):
        if self.is_min_id_length_valid() and self.min_id_length:
            return {"min_id_length": self.min_id_length}
        return None



from fga.event import BaseEvent
import pydash


class UserEvent:

    @staticmethod
    def __mapping__(event: BaseEvent):
        user_event = {"event_id": event.event_id, "user_id": event.user_id, "project_id": None, "timestamp": event.time,
                      "device_id": event.device_id or pydash.get(event, "event_properties.device_id"),
                      "region": event.region or pydash.get(event, "event_properties.region"),
                      "event_device": event.platform or pydash.get(event, "event_properties.platform"), "event_source": "web2",
                      "event_type": event.event_type, "extra_data": event}
        return user_event


class User:

    @staticmethod
    def __mapping__(event: BaseEvent):
        user = {
            "user_id": event.user_id,
            "project_id": None,
            "sign_up_device": pydash.get(event, "user_properties.sign_up_device"),
            "account_type": None,
            "idfa": event.idfa or event.idfv or pydash.get(event, "user_properties.idfa"),
            "udid": event.android_id or pydash.get(event, "user_properties.android_id"),
            "country": event.country or pydash.get(event, "user_properties.country"),
            "email": pydash.get(event, "user_properties.email"),
            "ip": event.ip or pydash.get(event, "user_properties.ip"),
            "twitter": pydash.get(event, "user_properties.twitter"),
            "discord": pydash.get(event, "user_properties.discord"),
            "sign_up_at": pydash.get(event, "user_properties.sign_up_at") or event.time,
            "extra_data": event
        }
        return user



import copy
import logging
import threading

from fga.constants import PluginType
from fga.exception import InvalidEventError
from fga.constants import LOGGER_NAME


class Timeline:

    def __init__(self, configuration=None):
        self.locks = {
            PluginType.BEFORE: threading.Lock(),
            PluginType.ENRICHMENT: threading.Lock(),
            PluginType.DESTINATION: threading.Lock()
        }
        self.plugins = {
            PluginType.BEFORE: [],
            PluginType.ENRICHMENT: [],
            PluginType.DESTINATION: []
        }
        self.configuration = configuration

    @property
    def logger(self):
        if self.configuration:
            return self.configuration.logger
        return logging.getLogger(LOGGER_NAME)

    def setup(self, client):
        self.configuration = client.configuration

    def add(self, plugin):
        with self.locks[plugin.plugin_type]:
            self.plugins[plugin.plugin_type].append(plugin)

    def remove(self, plugin):
        for plugin_type in self.locks:
            with self.locks[plugin_type]:
                self.plugins[plugin_type] = [p for p in self.plugins[plugin_type] if p != plugin]

    def flush(self):
        destination_futures = []
        for destination in self.plugins[PluginType.DESTINATION]:
            try:
                destination_futures.append(destination.flush())
            except Exception:
                self.logger.exception("Error for flush events")
        return destination_futures

    def process(self, event):
        if self.configuration.opt_out:
            self.logger.info("Skipped event for opt out config")
            return event
        before_result = self.apply_plugins(PluginType.BEFORE, event)
        enrich_result = self.apply_plugins(PluginType.ENRICHMENT, before_result)
        self.apply_plugins(PluginType.DESTINATION, enrich_result)
        return enrich_result

    def apply_plugins(self, plugin_type, event):
        result = event
        with self.locks[plugin_type]:
            for plugin in self.plugins[plugin_type]:
                if not result:
                    break
                try:
                    if plugin.plugin_type == PluginType.DESTINATION:
                        print('plugin_type', plugin.plugin_type)
                        plugin.execute(copy.deepcopy(result))
                    else:
                        result = plugin.execute(result)
                except InvalidEventError as e:
                    print("InvalidEventError", e)
                    self.logger.exception(f"Invalid event body {event}")
                except Exception as e:
                    print("Exception", e)
                    self.logger.exception(f"Error for apply {plugin_type.name} plugin for event {event}")
        return result

    def shutdown(self):
        for destination in self.plugins[PluginType.DESTINATION]:
            destination.shutdown()



import abc
from threading import Condition
from typing import List, Tuple, Optional

from fga.event import BaseEvent
from fga import utils, constants


class Storage(abc.ABC):

    @abc.abstractmethod
    def __init__(self):
        pass

    @abc.abstractmethod
    def push(self, event: BaseEvent) -> None:
        pass

    @abc.abstractmethod
    def pull(self, batch_size: int) -> List[BaseEvent]:
        pass

    @abc.abstractmethod
    def pull_all(self) -> List[BaseEvent]:
        pass


class StorageProvider(abc.ABC):

    @abc.abstractmethod
    def get_storage(self) -> Storage:
        pass


class InMemoryStorage(Storage):
    def __init__(self):
        self.total_events: int = 0
        self.buffer_data: List[Tuple[int, BaseEvent]] = []
        self.ready_queue: List[BaseEvent] = []
        self.buffer_lock_cv = Condition()
        self.configuration = None
        self.workers = None

    @property
    def lock(self):
        return self.buffer_lock_cv

    @property
    def max_retry(self) -> int:
        return self.configuration.flush_max_retries

    @property
    def wait_time(self) -> int:
        if self.ready_queue:
            return 0
        if self.buffer_data:
            return min(self.buffer_data[0][0] - utils.current_milliseconds(), self.configuration.flush_interval_millis)
        return self.configuration.flush_interval_millis

    def setup(self, configuration, workers):
        self.configuration = configuration
        self.workers = workers

    def push(self, event: BaseEvent, delay: int = 0) -> Tuple[bool, Optional[str]]:
        if event.retry and self.total_events >= constants.MAX_BUFFER_CAPACITY:
            return False, "Destination buffer full. Retry temporarily disabled"
        if event.retry >= self.max_retry:
            return False, f"Event reached max retry times {self.max_retry}."
        total_delay = delay + self._get_retry_delay(event.retry)
        self._insert_event(total_delay, event)
        self.workers.start()
        return True, None

    def pull(self, batch_size: int) -> List[BaseEvent]:
        current_time = utils.current_milliseconds()
        with self.lock:
            result = self.ready_queue[:batch_size]
            self.ready_queue = self.ready_queue[batch_size:]
            index = 0
            while index < len(self.buffer_data) and index < (batch_size - len(result)) and \
                    current_time >= self.buffer_data[index][0]:
                event = self.buffer_data[index][1]
                result.append(event)
                index += 1
            self.buffer_data = self.buffer_data[index:]
            self.total_events -= len(result)
        return result

    def pull_all(self) -> List[BaseEvent]:
        with self.lock:
            self.total_events = 0
            result = self.ready_queue + [element[1] for element in self.buffer_data]
            self.buffer_data = []
            self.ready_queue = []
            return result

    def _insert_event(self, total_delay: int, event: BaseEvent):
        current_time = utils.current_milliseconds()
        with self.lock:
            while self.buffer_data and self.buffer_data[0][0] <= current_time:
                self.ready_queue.append(self.buffer_data.pop(0)[1])
            if total_delay == 0:
                self.ready_queue.append(event)
            else:
                time_stamp = current_time + total_delay
                left, right = 0, len(self.buffer_data) - 1
                while left <= right:
                    mid = (left + right) // 2
                    if self.buffer_data[mid][0] > time_stamp:
                        right = mid - 1
                    else:
                        left = mid + 1
                self.buffer_data.insert(left, (time_stamp, event))
            self.total_events += 1
            if len(self.ready_queue) >= self.configuration.flush_queue_size:
                self.lock.notify()

    def _get_retry_delay(self, retry: int) -> int:
        if retry > self.configuration.flush_max_retries:
            return 3200
        if retry <= 0:
            return 0
        return 100 * (2 ** ((retry - 1) // 2))


class InMemoryStorageProvider(StorageProvider):

    def get_storage(self) -> InMemoryStorage:
        return InMemoryStorage()
